#!/usr/bin/python
# -*- coding: utf-8 -*-
#"""
#KOPR 1.0
#"""

import alb3
import comephem
from datavar import Colorlist
import datetime
import koprhand
import koprclass
import koprdial
import npfunc
from math import log10, ceil, pow
import numpy as np
from PyQt5 import QtCore, QtGui, QtWidgets
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.widgets import Slider, Button, RadioButtons
from matplotlib.figure import Figure

# Output is list of data in format [ colour, comet, magarams, returns ]
#                                      0        1      2        3         4
#                                    str     str      list      list
# Magparams is list in format [ h0,  n, Tbeg, Tend ]
#                                0   1    2     3
#                             float float str str
# Returns  is list in format [ Tyear, peridict,  Obs ]
# Obs is list in format [ otype, obs, uncsign, negsign,  m1, m1dist, phase, fwsca, date, deltaT, r, logr ]
#                           0     1      2       3       4      5      6      7      8     9     10 11 
#                          str  str     str     str    float float  float  float   ???    float float float
#                         filt  filt    filt   filt      y1  ycorr1 ycorr2 ycorr3   x1     x2    x3    x4     

class LightCurve(FigureCanvas):

    def __init__(self, parent=None, width=5, height=4, dpi=100 ):
        self.fig = Figure(figsize=(width, height), dpi=dpi, facecolor='lightgray')
        FigureCanvas.__init__(self, self.fig)
        self.setParent(parent)
        FigureCanvas.setSizePolicy(self, QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        FigureCanvas.updateGeometry(self)

    def setaxes1(self):
        self.ax = self.fig.add_subplot(111, facecolor='lightgray')
#        self.fig.tight_layout()
        self.ax.grid(True)
        self.canvas = self.ax.get_figure().canvas
        self.canvas.setFocusPolicy( QtCore.Qt.ClickFocus )
        self.canvas.setFocus()
        
    def setaxes2(self):
        self.ax1 = self.fig.add_subplot(121, facecolor='lightgray')
        self.ax2 = self.fig.add_subplot(122, facecolor='lightgray')
        self.fig.subplots_adjust(wspace=0, hspace=0)
        self.ax1.grid(True)
        self.ax2.grid(True)        
        self.ax1.invert_xaxis()
        self.ax2.yaxis.tick_right()
        self.canvas = self.ax1.get_figure().canvas
        self.canvas = self.ax2.get_figure().canvas
        self.canvas.setFocusPolicy( QtCore.Qt.ClickFocus )
        self.canvas.setFocus()

def LoadMultipleData(script, settings):
    data, icqlines, observers, text = [], [], [], []
    colors = ['black'] + Colorlist()

    for line in script:
      line = line.split(';')
      destig = line[0]
      color =  colors[int(line[1])]
      selection = GetType(line[2])
      yearsel = line[3].split('/')
      # open file of comet
      kometa = koprclass.comet()
      kometa.LoadComet(destig)     
      elements = koprhand.loadcometid(destig)
      
      text.append(destig)
      magparams = []
      params = line[4].split('/')
      for item in params:
        text.append(item)
        h0, rest = item.split(',',1)
        n, rest = rest.split('[')
        beg, end = rest.replace(']','').split(',')
        magparams.append( [ float(h0.replace('H0=','').replace(' mag','')), float(n.replace('n=','')), beg, end ] )

      if len(magparams) == 0: 
        rest = elements[90:102].split()
        magparams.append( [float(rest[0]), float(rest[1]), 'inf', 'inf'] )

      try:
        icqlines = koprhand.LoadValidObsList(destig)
      except IOError:            
         koprdial.InfoDialog(1,'Local data missing for '+destig,'Load data from COBS or MPC first')

      tyears = []
      for tdate in kometa.perihelions:
        tyears.append( tdate[0:4] )
        
      if len(yearsel[0]) == 0:
        yearsel = tyears

      pjdays, peridict = [], {}
      for item in kometa.perihelions:
        tjday = comephem.eDatetoJday(item)
        pjdays.append(tjday)
        peridict[tjday] = [ item.split('/')[0], kometa.elements[item] ]       
      pjdays = np.array(pjdays)         
      
      obslist = []
      for item in icqlines:
        obs = koprclass.icqcode()
        obs.decode_icq(item)
        if obs.otype in selection:
          jday = comephem.eDatetoJday(obs.date)
          tjday = pjdays[(np.abs(pjdays-jday)).argmin()] # find nearest perihelion date
          year = peridict[tjday][0]

          calc = comephem.ephem()
          calc.readdb(peridict[tjday][1])
          calc.compute(obs.date, '00:00', ['comet', 'extend'])
            
          if obs.obscode not in [x[0] for x in observers]:
            observers.append( [obs.obscode, 1] )

          outline = [ obs.otype, obs.obscode, obs.uncsign, obs.negsign, 
                      obs.mag, - 5*log10(calc.dist), alb3.phasemag(calc.phase),  - alb3.fwscmag(calc.phase,float(settings),0.9,-0.6,0.95),
                      comephem.JdaytoDtime(jday) , jday-tjday, calc.r, log10(calc.r) ]
          obslist.append( [ year, outline ] )
                               
      returns = []
      for year in yearsel:
        returns.append( [year, peridict, [ x[1] for x in obslist if x[0] == year ] ] )

      data.append( [ color, destig, magparams, returns ] )
      
    return data, observers, text

def GetType(codes):
    items = [*codes]
    sel = []
    for item in items:
      if item == 'v':
        sel.append('vis')
      elif item == 'c':
        sel.append('ccd')
      elif item == 'm':
        sel.append('mpct')
      elif item == 'n':
        sel.append('mpcn')
    return sel
    
def magparams(magparams):
    mp_arrray = []
    for item in magparams:
      mp_arrray.append( [ item[0], item[1], float(item[2].replace('inf','-9999')), float(item[3].replace('inf','9999'))] )
    return mp_arrray
    
def DtimeToStr(dtime):
    return dtime.strftime('%Y/%m/%d') + str(round( dtime.hour/24. + dtime.minute/1440. + dtime.second/86400., 2 ))[1:]

def DtimetoJday(dtime):
    return comephem.eDatetoJday(DtimeToStr(dtime))

def YMDtoDtime(y,m,d):
    y, m, d = int(y), int(m), float(d)
    df = d - int(d)
    d = d - df
    h = df * 24.
    hf = h - int(h)
    n = hf * 60.
    nf = n - int(n)
    s = nf * 60.

    return datetime.datetime(int(y), int(m), int(d), int(h), int(n), int(s) )

def StrToDtime(text):
    y, m, d = text.split('/')
    
    return YMDtoDtime(y,m,d)
    
def RoundDates( xmin, xmax ):
    xmin = xmin.replace(microsecond=0, second=0, minute=0, hour=0)
    xmax = xmax.replace(microsecond=0, second=0, minute=0, hour=0)+ datetime.timedelta( days = 1 )
    
    return xmin, xmax

def HalfW(eline):
    calc = comephem.ephem()
    calc.readdb(eline)

    return npfunc.perilen(calc.e, calc.q)

def halfwnorm(param, halfw):
    if param[2] < -halfw: # normalisation of infinite elements to actual range
      param[2] = -halfw
    if param[3] > halfw: # normalisation of infinite elements to actual range
      param[3] = halfw
    
    return param

def qRangNorm(q, rang):
    if rang < q:
      return q
    else:
      return rang

# [type, beg, end], T datetime, q, e
def GetRange(ranges, tdtime, q, e):
    if ranges[0] == 0:
      range1 = ( ranges[1] - tdtime).total_seconds()/86400.
      range2 = ( ranges[2] - tdtime).total_seconds()/86400.
    elif ranges[0] == 1:
      range1 = ranges[1]
      range2 = ranges[2] 
    elif ranges[0] == 2:
      range1 = alb3.finddt( qRangNorm(q, ranges[1]),q,e)
      range2 = alb3.finddt( qRangNorm(q, ranges[2]),q,e)
    elif ranges[0] == 3:
      range1 = alb3.finddt( qRangNorm(q, pow( 10., ranges[1])),q,e)
      range2 = alb3.finddt( qRangNorm(q, pow( 10., ranges[2])),q,e)
    return range1, range2 # returns a range in delta days

# points list in format [ rmag, dist, phase, fwsca, date, deltaT, r, logr ]
#                         0     1       2      3     4      5     6   7 
#                       float float  float  float   ???    float float float
#                         y1  ycorr1 ycorr2 ycorr3   x1     x2    x3    x4           
def GeneratePoints(eline, param, set_filter, settings):
    points = []
    calc = comephem.ephem()
    calc.readdb(eline)
    tdtime = YMDtoDtime( calc.tyear , calc.tmonth , calc.tday )
    range1, range2 = GetRange(set_filter, tdtime, calc.q, calc.e)
    dt_beg, dt_end = param[2], param[3]

    if set_filter[0] > 1:
      if dt_beg < 0 and dt_end <= 0 :
        if dt_beg < - range2:
          dt_beg = - range2
        if dt_end > - range1:
          dt_end = - range1
      elif dt_beg < 0 and dt_end > 0 :
        if dt_beg < - range2:
          dt_beg = - range2
        if dt_end > range2:
          dt_end = range2
      elif dt_beg >= 0 and dt_end > 0 :
        if dt_beg < range1:
          dt_beg = range1
        if dt_end > range2:
          dt_end = range2
    else:  
      if dt_beg < range1:
        dt_beg = range1
      if dt_end > range2:
        dt_end = range2
        
    step = dt_end - dt_beg  # number of steps in days
    if step > 0:   
      start = tdtime + datetime.timedelta( days = dt_beg ) # start date in datetime
      for i in range(ceil(step)):
        date = start + datetime.timedelta( days = i )
        calc.compute(koprhand.cvt2time( date ), '00:00', ['comet', 'extend'])
        points.append( [ param[0] + 2.5*param[1]*log10(calc.r) ,                   # dist corrected mag
                         5 * log10(calc.dist),                                     # distance mag correction
                         alb3.phasemag(calc.phase),                                # phase correction
                         - alb3.fwscmag(calc.phase,float(settings),0.9,-0.6,0.95), # forward scattering correction
                         date,                                                     # date
                         dt_beg + float(i),                                        # delta t
                         calc.r,                                                   # r
                         log10(calc.r) ] )                                         # logr
      
    return points  

def xypoints(data, lim, i, j, k, l, typ, obs):
    pret, post = [x for x in data if x[9] < 0], [x for x in data if x[9] >= 0]
        
    if len(lim) == 4:
      xpoints1 = [ [ x[i], x[0], x[1] ] for x in pret if lim[0] <= x[i] <= lim[1] and lim[3] <= x[4] + j * x[5] + k * x[6] + l * x[7] <= lim[2] ]
      ypoints1 = [ [ x[4] + j * x[5] + k * x[6] + l * x[7], x[0], x[1] ]
                  for x in pret 
                  if lim[0] <= x[i] <= lim[1] and lim[3] <= x[4] + j * x[5] + k * x[6] + l * x[7] <= lim[2] ]
      xpoints2 = [ [ x[i], x[0], x[1] ] for x in post if lim[0] <= x[i] <= lim[1] and lim[3] <= x[4] + j * x[5] + k * x[6] + l * x[7] <= lim[2] ]
      ypoints2 = [ [ x[4] + j * x[5] + k * x[6] + l * x[7], x[0], x[1] ]
                  for x in post 
                  if lim[0] <= x[i] <= lim[1] and lim[3] <= x[4] + j * x[5] + k * x[6] + l * x[7] <= lim[2] ]
    else:
      xpoints1 = [ [ x[i], x[0], x[1] ] for x in pret]
      ypoints1 = [ [ x[4] + j * x[5] + k * x[6] + l * x[7], x[0], x[1] ] for x in pret]
      xpoints2 = [ [ x[i], x[0], x[1] ] for x in post]
      ypoints2 = [ [ x[4] + j * x[5] + k * x[6] + l * x[7], x[0], x[1] ] for x in post]

    if typ == 1:
      types = [x[0] for x in self.types if x[1] == 1 ]
      xpoints1, ypoints1 = [x[0] for x in xpoints1 if x[1] in types ], [x[0] for x in ypoints1 if x[1] in types ]
      xpoints2, ypoints2 = [x[0] for x in xpoints2 if x[1] in types ], [x[0] for x in ypoints2 if x[1] in types ]

    if obs == 1:
      obscode = [x[0] for x in self.obscode if x[1] == 1 ]
      xpoints1, ypoints1 = [x[0] for x in xpoints1 if x[2] in obscode], [x[0] for x in ypoints1 if x[2] in obscode]
      xpoints2, ypoints2 = [x[0] for x in xpoints2 if x[2] in obscode], [x[0] for x in ypoints2 if x[2] in obscode]
        
    if obs == 0 and typ == 0:
      xpoints1, ypoints1 = [x[0] for x in xpoints1], [x[0] for x in ypoints1]
      xpoints2, ypoints2 = [x[0] for x in xpoints2], [x[0] for x in ypoints2]    
          
    return xpoints1, ypoints1, xpoints2, ypoints2
